from .on_connect import on_connect
from .on_connect_fail import on_connect_fail
from .on_message import on_message
from .on_subscribe import on_subscribe

__all__ = [on_message, on_connect, on_connect_fail, on_subscribe]



from pathlib import Path
from typing import Union

import paho.mqtt.client

import balance_subscriber.callbacks


def get_client(
    topics: set[str],
    data_dir: Union[str, Path],
    encoding: str = None,
    username: str = None,
    password: str = None,
    qos: int = 0,
    ext: str = None,
) -> paho.mqtt.client.Client:
    """
    Initialise the MQTT client
    """
    if not data_dir:
        raise ValueError("No data directory specified")

    encoding = encoding or "utf-8"
    ext = ext or ".bin"

    # Initialise client
    client = paho.mqtt.client.Client(paho.mqtt.client.CallbackAPIVersion.VERSION2)
    # https://eclipse.dev/paho/files/paho.mqtt.python/html/index.html#logger
    client.enable_logger()
    # Make the topics available to the on_connect callback
    client.user_data_set(
        dict(
            topics=topics, data_dir=Path(data_dir), encoding=encoding, qos=qos, ext=ext
        )
    )

    # Authentication
    if username:
        client.username_pw_set(username=username, password=password)

    # Register callbacks
    client.on_connect = balance_subscriber.callbacks.on_connect
    client.on_connect_fail = balance_subscriber.callbacks.on_connect_fail
    client.on_message = balance_subscriber.callbacks.on_message
    client.on_subscribe = balance_subscriber.callbacks.on_subscribe

    return client



import csv
import datetime
import logging
import tempfile
import time
from pathlib import Path

import paho.mqtt.client
import pytest
import balance_subscriber.callbacks

logger = logging.getLogger(__name__)


@pytest.mark.parametrize("payload", ["471.22", 471.22, "disabled", 123, "192.168.1.199"])
def test_on_message(payload):
    # Message options
    encoding = "utf-8"
    userdata = dict(data_dir=tempfile.mkdtemp(), encoding=encoding)
    topic = "plant/PL-f15320/Loadcell-B"

    # Build MQTT message
    msg = paho.mqtt.client.MQTTMessage(mid=0, topic=topic.encode(encoding))
    msg.payload = str(payload).encode(encoding)
    msg.timestamp = time.monotonic()

    # Run callback
    balance_subscriber.callbacks.on_message(None, userdata, msg)

    # Check data has been serialised to CSV
    path = Path(userdata["data_dir"]).joinpath(f"{topic}.csv")
    with path.open() as file:
        reader = csv.reader(file)
        row = next(reader)
        timestamp, value = row

        # Check data types
        datetime.datetime.fromisoformat(timestamp)
        assert isinstance(value, str)
        assert value == str(payload)



import tempfile
import uuid

from balance_subscriber.client import get_client

HOSTNAME = "test.mosquitto.org"


def test_publish():
    topic = uuid.uuid4().hex
    with tempfile.NamedTemporaryFile() as temp:
        client = get_client(topics={topic}, data_dir=temp.name)
        client.connect(host=HOSTNAME)
        client.publish(topic=topic, payload="Hello world")
        client.disconnect()



import tempfile

import balance_subscriber.client


def test_get_client():
    balance_subscriber.client.get_client(
        topics={"topic1/subtopic1", "topic1/subtopic2", "topic2/subtopic1"},
        data_dir=tempfile.mkdtemp(),
        encoding="utf-8",
    )



import argparse
import importlib.metadata
import logging
import os
from pathlib import Path

from balance_subscriber.client import get_client

logger = logging.getLogger(__name__)

DESCRIPTION = """
This is an MQTT subscriber that serialises incoming messages.
"""


def get_args():
    """
    Command-line arguments
    """
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        "--data_dir",
        "-d",
        type=Path,
        help="Directory to save messages to.",
        default=os.getenv("DATA_DIR"),
    )
    parser.add_argument(
        "--log_level",
        default=os.getenv("LOG_LEVEL", "WARNING"),
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )
    parser.add_argument(
        "topics",
        nargs="*",
        help="Topics to subscribe to, default: all",
        default=os.getenv("TOPICS", "#"),
    )
    parser.add_argument(
        "--host",
        type=str,
        default=os.getenv("HOST", "localhost"),
        help="MQTT broker host",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=int(os.getenv("PORT", 1883)),
        help="MQTT broker port",
    )
    parser.add_argument(
        "--keepalive",
        type=int,
        default=int(os.getenv("KEEP_ALIVE", 60)),
        help="MQTT broker keep-alive interval",
    )
    parser.add_argument(
        "--encoding", default="utf-8", help="CSV output text character set"
    )
    parser.add_argument(
        "--version",
        action="version",
        version=importlib.metadata.version("balance-subscriber"),
    )
    parser.add_argument("--username", "-u", default=os.getenv("USERNAME"))
    parser.add_argument("--password", "-p", default=os.getenv("PASSWORD"))
    parser.add_argument(
        "-q", "--qos", type=int, choices={0, 1, 2}, default=0, help="Quality of service"
    )
    parser.add_argument(
        "-e",
        "--ext",
        default=os.getenv("EXT", ".bin"),
        help="File extension for saved data.",
    )
    return parser.parse_args()


def main():
    args = get_args()
    logging.basicConfig(level=args.log_level)

    # Connect to message broker
    client = get_client(
        topics=args.topics,
        data_dir=args.data_dir,
        encoding=args.encoding,
        username=args.username,
        password=args.password,
        qos=args.qos,
        ext=args.ext,
    )
    client.connect(host=args.host, port=args.port, keepalive=args.keepalive)

    # Blocking call that processes network traffic, dispatches callbacks and handles reconnecting.
    try:
        client.loop_forever()
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



import csv
import datetime
import logging
from pathlib import Path

from paho.mqtt.client import Client, MQTTMessage

logger = logging.getLogger(__name__)


def on_message(client: Client, userdata: dict, msg: MQTTMessage):
    """
    The callback for when a PUBLISH message is received from the server.

    on_message callback
    https://eclipse.dev/paho/files/paho.mqtt.python/html/client.html#paho.mqtt.client.Client.on_message

    MQTT message class
    https://eclipse.dev/paho/files/paho.mqtt.python/html/client.html#paho.mqtt.client.MQTTMessage
    """

    ext = userdata.get("ext", ".csv")
    "File extension"
    encoding = userdata.get("encoding", "utf-8")
    data_dir = Path(userdata["data_dir"])
    "The root directory for serialising messages."

    # Convert an MQTT topic to a file path
    # E.g. 'plant/PL-f15320/Network' becomes 'plant/PL-f15320/Network.csv'
    path = data_dir / f"{msg.topic}{ext}"
    # Ensure subdirectory exists
    path.parent.mkdir(parents=True, exist_ok=True)

    # Build data row
    row = (
        # Current timestamp
        datetime.datetime.now(datetime.UTC).isoformat(),
        msg.payload.decode(encoding),
    )

    # Serialise payload by appending a row to the CSV file
    with path.open(mode="a", encoding=encoding) as file:
        writer = csv.writer(file)
        writer.writerow(row)



import logging

logger = logging.getLogger(__name__)


def on_connect_fail(client, userdata):
    """
    The callback called when the client failed to connect to the broker.
    https://eclipse.dev/paho/files/paho.mqtt.python/html/client.html#paho.mqtt.client.Client.on_connect_fail

    Called by loop_forever() and loop_start() when the TCP connection failed to establish. This callback is not called
    when using connect() or reconnect() directly. It’s only called following an automatic (re)connection made by
    loop_start() and loop_forever()
    """
    logger.error("on_connect_fail")
    raise ConnectionRefusedError("on_connect_fail")



import tempfile
from paho.mqtt.reasoncodes import ReasonCode
from paho.mqtt.packettypes import PacketTypes
import balance_subscriber.callbacks
import balance_subscriber.client


def test_on_connect():
    topics = {"topic1/subtopic1", "topic1/subtopic2", "topic2/subtopic1"}
    client = balance_subscriber.client.get_client(
        topics=topics, data_dir=tempfile.mkdtemp(), encoding="utf-8"
    )
    userdata = dict(topics=topics)
    reason_code = ReasonCode(PacketTypes.DISCONNECT)
    balance_subscriber.callbacks.on_connect(client, userdata, None, reason_code, None)



import logging

from paho.mqtt.reasoncodes import ReasonCode

from balance_subscriber.exceptions import SubscriptionError

logger = logging.getLogger(__name__)


def on_subscribe(client, userdata, mid, reason_code_list: list[ReasonCode], properties):
    """
    This callback is called when the broker responds to a subscribe request, sending a SUBACK response.
    https://eclipse.dev/paho/files/paho.mqtt.python/html/client.html#paho.mqtt.client.Client.on_subscribe
    """

    for reason_code in reason_code_list:
        if reason_code.is_failure:
            raise SubscriptionError(reason_code)
        logger.info("Subscribed. Reason code '%s'", reason_code)



import logging

from paho.mqtt.reasoncodes import ReasonCode

logger = logging.getLogger(__name__)


def on_connect(client, userdata, connect_flags, reason_code: ReasonCode, properties):
    """
    The callback for when the client receives a CONNACK response from the server.
    https://eclipse.dev/paho/files/paho.mqtt.python/html/client.html#paho.mqtt.client.Client.on_connect
    """
    logger.info("Reason code: '%s'", reason_code)

    if reason_code.is_failure:
        raise ConnectionError(reason_code)

    # Quality of service
    qos = userdata.get("qos", 0)

    # Subscribing in on_connect() means that if we lose the connection and
    # reconnect then subscriptions will be renewed.
    for topic in userdata["topics"]:
        logger.info("Attempting to subscribe to topic '%s'...", topic)
        client.subscribe(topic, qos=qos)
